import logging, os, unittest
from dataclasses import dataclass, field

from lilvali import validator, validate
from lilvali.errors import *


log = logging.getLogger(__name__)

if os.environ.get("LILVALI_DEBUG", False) == "True":  # pragma: no cover
    logging.basicConfig(
        level=logging.DEBUG, format="%(funcName)s %(name)s:%(lineno)s %(message)s"
    )


@validate
@dataclass
class SomeClass:
    x: int
    y: str = field(default="hello")

    @validator
    def _x(value):
        if value is None or value < 0:
            raise ValidationError

    @validator
    def _y(value) -> bool:
        return value == "hello"


@validate
class NotADC:
    def __init__(self, x: int, y: str):
        self.x = x
        self.y = y

    @validator
    def _x(value):
        if value is None or value < 0:
            raise ValidationError

    @validator
    def _y(value) -> bool:
        if value != "hello":
            raise ValidationError


class TestValidateTypes(unittest.TestCase):
    def test_dataclass(self):
        self.assertEqual(SomeClass(1, "hello").x, 1)
        self.assertEqual(SomeClass(1).y, "hello")
        with self.assertRaises(ValidationError):
            SomeClass(-1.3, "hello")
        with self.assertRaises(ValidationError):
            SomeClass(1, None)
        with self.assertRaises(ValidationError):
            SomeClass(1.0, "hello")
        with self.assertRaises(ValidationError):
            SomeClass(-1, "hello")
        with self.assertRaises(ValidationError):
            SomeClass(1, "herro")

    def test_not_dataclass(self):
        self.assertEqual(NotADC(1, "hello").x, 1)


def main():
    import pdb

    pdb.set_trace()
    SomeClass(1, "hello")


if __name__ == "__main__":
    main()



""" unittest to check validation of NewTypes, which should just be the same as the basic types they are based on."""
import logging, os, unittest
from typing import NewType

from lilvali import validate
from lilvali.errors import *


class TestValidateNewTypes(unittest.TestCase):
    def test_validate_newtype(self):
        """Test that validation of newtypes is the same as the base type."""

        @validate
        def validate_int(value: int):
            return value > 0

        NT = NewType("Int", int)

        @validate
        def validate_newtype(value: NT):
            return value > 0

        self.assertEqual(validate_int(1), validate_newtype(1))

        with self.assertRaises(ValidationError):
            validate_newtype(1.0)



import inspect
import json, os, logging
from dataclasses import dataclass
from dataclass_wizard import JSONWizard
import dataclass_wizard

from lilvali import validate, validator
from lilvali.errors import *


log = logging.getLogger(__name__)
if os.environ.get("LILVALI_DEBUG", False) == "True":  # pragma: no cover
    logging.basicConfig(level=logging.DEBUG, format="%(name)s:%(lineno)s %(message)s")


def damn(decorator):
    def decorate(cls):
        for attr_name, attr_value in cls.__dict__.items():
            if not attr_name.startswith("_") and callable(func := attr_value):
                log.debug(f"decorating {attr_name} with {decorator}")
                func.validated_base_cls = cls
                setattr(cls, attr_name, decorator(func))
        return cls

    return decorate


def test_validate():
    @validate
    def add[T: (int, float)](x: int, y: T) -> int | float:
        return x + y

    print(f"{add(1, 2)=}")
    print(f"{add(1, 2.0)=}")

    try:
        print(f"{add(1.0, 2)=}")
    except ValidationError as e:
        pass


def test_validated_cls():
    total_validate = lambda cls: damn(validate)(validate(cls))

    @total_validate
    @dataclass
    class SomeSchema(JSONWizard):
        """a dataclass that defines a json schema."""

        name: str
        data: dict[str, int]

        @classmethod
        def from_dict(cls, value: dict):
            return super().from_dict(value)

    c = SomeSchema.from_dict({"name": "A", "data": {}})
    print(c)

    try:
        SomeSchema.from_dict("not_a_dict")
    except ValidationError:
        print("GOTCHA!")
    except dataclass_wizard.errors.MissingFields:
        print("DARN IT WE LOST 'ER!")

    try:
        SomeSchema(1, False)
    except ValidationError:
        print("GOOD STUFF!")


def main():
    test_validate()
    test_validated_cls()


if __name__ == "__main__":
    main()



from .checker import ValidationBindChecker, ValidatorFunction
from .validator import TypeValidator
from .decorators import validate, validator


__all__ = [
    "validate",
    "validator",
    "TypeValidator",
    "ValidatorFunction",
    "ValidationBindChecker",
]



import inspect, logging
from itertools import chain
from typing import (
    Callable,
)


from .checker import ValidationBindChecker


log = logging.getLogger(__name__)


class TypeValidator:
    """Callable wrapper for validating function arguments and return values."""

    def __init__(self, func: type | Callable, config=None):
        self.func = func
        if hasattr(func, "validated_base_cls"):
            self._cls = func.validated_base_cls
        else:
            log.debug("NO CLASS!")
            self._cls = None

        self.argspec, self.generics = inspect.getfullargspec(func), func.__type_params__

        self.bind_checker = ValidationBindChecker(config=config)

    def __call__(self, *args, **kwargs):
        """Validating wrapper for the bound self.func"""
        # if the function is a method, add the class to the args.
        if "self" in self.argspec.args:
            if self._cls is not None:
                args = (self._cls, *args)

        # If disabled, just call the function being validated.
        if self.bind_checker.config.disabled:
            return self.func(*args, **kwargs)

        # First refresh the BindChecker with new bindings on func call,
        self.bind_checker.new_bindings(self.generics)

        fixed_args = zip(self.argspec.args, args)
        var_args = (
            (self.argspec.varargs, arg) for arg in args[len(self.argspec.args) :]
        )
        all_args = list(chain(fixed_args, var_args, kwargs.items()))

        # then check all args against their type hints.
        for name, arg in all_args:
            ann = self.argspec.annotations.get(name)
            if ann is not None:
                # print(f"{name=} {ann=} {arg=}")
                self.bind_checker.check(ann, arg)

        # After ensuring all generic values can bind,
        checked = self.bind_checker.checked
        if all(checked):
            # call the function being validated.
            result = self.func(*args, **kwargs)

            # If there is a return annotation
            if "return" in self.argspec.annotations:
                ret_ann = self.argspec.annotations["return"]
                log.debug(
                    "Return: annotations=%s result_type=%s return_spec=%s",
                    self.argspec.annotations,
                    type(result),
                    ret_ann,
                )

                # check it.
                if self.bind_checker.config.ret_validation and ret_ann is not None:
                    self.bind_checker.check(ret_ann, result)
            # Finally, return the results if nothing has gone wrong.
            return result

    def checking_on(self):
        """Turn type validation on."""
        self.bind_checker.config.disabled = False

    def checking_off(self):
        """Turn type validation off."""
        self.bind_checker.config.disabled = True



import logging
import os
import unittest


from lilvali.validate import validate, validator
from lilvali.errors import *


if os.environ.get("LILVALI_DEBUG", False) == "True":  # pragma: no cover
    logging.basicConfig(level=logging.DEBUG, format="%(name)s:%(lineno)s %(message)s")


class TestValidationFunctions(unittest.TestCase):
    def test_mymod(self):
        @validate
        def mymod[T](a: T, b: T):
            return a % b

        self.assertEqual(mymod(10, 3), 10 % 3)
        self.assertEqual(mymod(10.0, 3.0), 10.0 % 3.0)
        with self.assertRaises(ValidationError):
            mymod(10, 3.0)

    def test_mysum(self):
        @validate
        def mysum[T, W: (int, float)](a: T, b: T) -> W:
            return float(a + b)

        @validate
        def badmysum[T, W: (int, float)](a: T, b: T) -> W:
            return str(a + b)

        self.assertEqual(mysum(1, 2), 3.0)
        with self.assertRaises(ValidationError):
            mysum(1, 2.0)
        with self.assertRaises(ValidationError):
            badmysum(1, 2)

        badmysum.bind_checker.config.ret_validation = False
        self.assertEqual(badmysum(1, 2), "3")

    def test_variadic_func(self):
        @validate
        def variadic_func[T, *Us](a: T, *args: *Us) -> (Us, T):
            if isinstance(args[0], str):
                return f"{a}".join(args), a
            return args, a

        self.assertEqual(variadic_func(1, 2, 3, 4), ((2, 3, 4), 1))
        self.assertEqual(variadic_func(1.0, 2.0, 3.0), ((2.0, 3.0), 1.0))
        self.assertEqual(variadic_func("a", "b", "c"), ("bac", "a"))
        self.assertEqual(variadic_func("a", 1, 2, 3), ((1, 2, 3), "a"))
        with self.assertRaises(ValidationError):
            variadic_func(1, 2, 3.0)

    def test_variadic_arguments(self):
        @validate(config={"implied_lambdas": True})
        def func(a: int, *args: int):
            return a + sum(args)

        self.assertEqual(func(1, 2, 3), 6)
        with self.assertRaises(ValidationError):
            func(1, 2, "3")

    def test_default_arg_func(self):
        @validate
        def default_arg_func[T, U: (int, float)](a: T, b: U = 10) -> U:
            return b

        self.assertEqual(default_arg_func("Hello"), 10)
        self.assertEqual(default_arg_func("Hello", 20), 20)
        with self.assertRaises(ValidationError):
            default_arg_func("Hello", "World")

    def test_generic_sequence_func(self):
        @validate
        def generic_sequence_func[T, U](a: (T, U), b: [T]) -> dict:
            return {"first": a, "second": b}

        self.assertEqual(
            generic_sequence_func((1, "a"), [1, 2, 3]),
            {"first": (1, "a"), "second": [1, 2, 3]},
        )
        self.assertEqual(
            generic_sequence_func(("a", 1.0), ["a", "b"]),
            {"first": ("a", 1.0), "second": ["a", "b"]},
        )
        with self.assertRaises(ValidationError):
            generic_sequence_func((1, "a"), [1, "b"])

        @validate
        def generic_tuple_return[T, U: (int, float)](a: T, b: U) -> (T, U):
            return a, b

        self.assertEqual(generic_tuple_return(1, 2), (1, 2))
        self.assertEqual(generic_tuple_return(1, 2.0), (1, 2.0))
        with self.assertRaises(ValidationError):
            generic_tuple_return(1, "a")

    def test_generic_tuple(self):
        @validate
        def generic_tuple_func[T, U: (int, float)](a: (T, U)) -> (T, U):
            return a

        self.assertEqual(generic_tuple_func((1, 2)), (1, 2))
        self.assertEqual(generic_tuple_func(("a", 2.0)), ("a", 2.0))
        with self.assertRaises(ValidationError):
            generic_tuple_func((1, "a"))

        @validate
        def generic_tuple_func2[
            T, U: (int, float)
        ](a: (int, str), b: (str, float)) -> (T, U):
            return a, b

        with self.assertRaises(ValidationError):
            generic_tuple_func2((1, "a"), ("b", 2.0))
        with self.assertRaises(ValidationError):
            generic_tuple_func2((1, "a"), ("b", "c"))

        # now one to that should pass and fail on return
        @validate
        def generic_tuple_func3[
            T, U: ((int, str), float)
        ](a: (int, str), b: (str, float)) -> (T, U):
            return b, a

        self.assertEqual(
            generic_tuple_func3((1, "a"), ("b", 2.0)), (("b", 2.0), (1, "a"))
        )
        with self.assertRaises(ValidationError):
            generic_tuple_func3((1, "a"), ("b", "c"))

        with self.assertRaises(ValidationError):
            generic_tuple_func3([1, "a"], ("b", 2.0))

    def test_generic_dict(self):
        @validate
        def generic_dict_func[T, U: (int, float)](a: dict[T, U]) -> dict[T, U]:
            return a

        self.assertEqual(generic_dict_func({1: 2, 3: 4}), {1: 2, 3: 4})
        self.assertEqual(generic_dict_func({"a": 2.0, "b": 4.0}), {"a": 2.0, "b": 4.0})
        with self.assertRaises(ValidationError):
            generic_dict_func({1: 2, "a": "b"})

        with self.assertRaises(ValidationError):
            generic_dict_func([1, 2, 3])

    def test_with_custom_validator(self):
        has_e = validator(lambda arg: True if "e" in arg else False)

        @validate
        def with_custom_validator(s: has_e):
            return s

        self.assertEqual(with_custom_validator("Hello"), "Hello")
        with self.assertRaises(ValidationError):
            with_custom_validator("World")

        @validator(base=int)
        def has_c_or_int(arg):
            return True if "c" in arg else False

        # has_c_or_int = validator(lambda arg: True if "c" in arg else False, base=int)

        @validate
        def with_custom_validator2(s: has_c_or_int):
            return s

        self.assertEqual(with_custom_validator2(10), 10)
        self.assertEqual(with_custom_validator2("has_c"), "has_c")
        with self.assertRaises(ValidationError):
            self.assertEqual(with_custom_validator2("Hello"), "Hello")

    def test_generic_union(self):
        @validate
        def generic_union[T: (str, bool)](a: int | T) -> int | T:
            return a

        self.assertEqual(generic_union(10), 10)
        self.assertEqual(generic_union("Hello"), "Hello")
        with self.assertRaises(ValidationError):
            generic_union(10.0)

    def test_generic_union_with_constraints(self):
        @validate
        def add[T: (int, float)](x: int, y: T) -> int | float:
            return x + y

        self.assertEqual(add(1, 2), 3)
        self.assertEqual(add(1, 2.0), 3.0)
        with self.assertRaises(ValidationError):
            add(1.0, 2)

    def test_validate_dict(self):
        @validate
        def func_a(a: dict):
            return ";".join(f"{k},{v}" for k, v in a.items())

        @validate
        def func_b(a: dict[str, int]) -> int:
            return sum(a.values())

        self.assertEqual(func_a({"a": 1, "b": 2}), "a,1;b,2")
        self.assertEqual(func_b({"a": 1, "b": 2}), 3)
        with self.assertRaises(ValidationError):
            func_b({"a": 1, "b": "c"})

    def test_none_values(self):
        @validate
        def none_func[T: (int, str)](a: T) -> T:
            return a

        with self.assertRaises(ValidationError):
            none_func(None)

        none_func.checking_off()

        self.assertEqual(none_func(None), None)

    def test_no_annotations(self):
        @validate
        def no_anno_func(a, b):
            return a + b

        no_anno_func(1, 2)

    def test_nested_generics(self):
        @validate
        def nested_func[T: (int, str)](a: dict[str, list[T]]) -> list[T]:
            return [item for sublist in a.values() for item in sublist]

        self.assertEqual(nested_func({"a": [1, 2], "b": [3, 4]}), [1, 2, 3, 4])
        self.assertEqual(
            nested_func({"a": ["1", "2"], "b": ["3", "4"]}), ["1", "2", "3", "4"]
        )
        # because already joined to one type.
        with self.assertRaises(ValidationError):
            self.assertEqual(nested_func({"a": [1, 2], "b": [3, "4"]}), [1, 2, 3, "4"])

        # TODO: nested validation of dict values
        with self.assertRaises(ValidationError):
            nested_func({"a": [1, 2], "b": [3, 5.0]})

        # Some arcane PEP suggests this is possible...
        # @validate
        # def nested_func2[T, U]((x, y: T), z: U) -> (U, T):
        #     pass

    def test_multiple_validators(self):
        is_even = validator(lambda arg: arg % 2 == 0)
        is_positive = validator(lambda arg: arg > 0)

        @validate
        def and_multi_validator_func(a: is_even & is_positive):
            return a

        self.assertEqual(and_multi_validator_func(4), 4)
        with self.assertRaises(ValidationError):
            and_multi_validator_func(3)
        with self.assertRaises(ValidationError):
            and_multi_validator_func(-4)

        @validate
        def or_multi_validator_func(a: is_even | is_positive):
            return a

        self.assertEqual(or_multi_validator_func(4), 4)
        self.assertEqual(or_multi_validator_func(3), 3)
        self.assertEqual(or_multi_validator_func(-4), -4)
        with self.assertRaises(ValidationError):
            or_multi_validator_func("Hello")
        with self.assertRaises(ValidationError):
            or_multi_validator_func(-3)

    def test_custom_error_message(self):
        is_even = validator(lambda arg: arg % 2 == 0, error="Not an even number!")

        @validate
        def custom_error_func(a: is_even):
            return a

        custom_error_func.checking_off()
        self.assertEqual(custom_error_func(3), 3)

        custom_error_func.checking_on()
        with self.assertRaisesRegex(ValidationError, "Not an even number!"):
            custom_error_func(3)



import logging, os, unittest
from typing import List, Union, Optional, Callable, Any, TypedDict, Literal

from lilvali import validate
from lilvali.errors import *


class TestValidateTypes(unittest.TestCase):
    def setUp(self):
        if os.environ.get("LILVALI_DEBUG", False) == "True":  # pragma: no cover
            logging.basicConfig(
                level=logging.DEBUG, format="%(name)s:%(lineno)s %(message)s"
            )

        super().setUp()

    def test_type(self):
        @validate
        def type_func(a: type) -> bool:
            return isinstance(a, type)

        self.assertTrue(type_func(int))
        with self.assertRaises(ValidationError):
            type_func("Not a type")

    def test_basic_types(self):
        @validate
        def func(a: int, b: str, c: float, d: bool):
            return f"{a}, {b}, {c}, {d}"

        self.assertEqual(func(1, "test", 3.14, True), "1, test, 3.14, True")
        with self.assertRaises(ValidationError):
            func("1", "test", 3.14, True)

    def test_none_not_allowed(self):
        @validate
        def func(a: int):
            return a

        self.assertEqual(func(10), 10)
        with self.assertRaises(ValidationError):
            func(None)

    def test_none_allowed(self):
        @validate
        def func(a: Optional[int]):
            return a

        self.assertIsNone(func(None))
        self.assertEqual(func(10), 10)
        with self.assertRaises(ValidationError):
            func("test")

    def test_optional_types(self):
        @validate
        def func(a: Optional[int]):
            return a

        self.assertIsNone(func(None))
        self.assertEqual(func(10), 10)
        with self.assertRaises(ValidationError):
            func("test")

    def test_union(self):
        @validate
        def union_func(a: Union[int, str, float]) -> str:
            return str(a)

        self.assertEqual(union_func(5), "5")
        self.assertEqual(union_func("Hello"), "Hello")
        self.assertEqual(union_func(5.5), "5.5")
        with self.assertRaises(ValidationError):
            union_func([])

        @validate
        def union_func2(a: str | int | float) -> str:
            return str(a)

        self.assertEqual(union_func2(5), "5")
        self.assertEqual(union_func2("Hello"), "Hello")
        self.assertEqual(union_func2(5.5), "5.5")
        with self.assertRaises(ValidationError):
            union_func2(set([1]))
        with self.assertRaises(ValidationError):
            union_func2([])
        with self.assertRaises(ValidationError):
            union_func2({})

    def test_union_types(self):
        @validate
        def func(a: Union[int, str]):
            return a

        self.assertEqual(func(10), 10)
        self.assertEqual(func("test"), "test")
        with self.assertRaises(ValidationError):
            func(3.14)

    def test_callable_types(self):
        @validate(config={"implied_lambdas": True})
        def func(a: Callable[[int], str]):
            return a(10)

        def func2(a: int) -> str:
            return str(a)

        self.assertEqual(func(func2), "10")
        self.assertEqual(func(lambda x: str(x)), "10")
        with self.assertRaises(ValidationError):
            func(10)

    def test_callable(self):
        @validate
        def callable_func(a: Callable[[int], str]) -> str:
            return a(5)

        def the_callable(x: int) -> str:
            return str(x)

        self.assertEqual(callable_func(the_callable), "5")

        with self.assertRaises(ValidationError):
            callable_func(lambda x: str(x))

        with self.assertRaises(ValidationError):
            callable_func("Not a function")

        @validate(config={"implied_lambdas": True})
        def callable_func2(a: Callable[[int], str]) -> str:
            return a(5)

        self.assertEqual(callable_func2(lambda x: str(x)), "5")

    def test_nested_collections(self):
        @validate
        def func(a: List[List[int]]):
            return a

        self.assertEqual(func([[1, 2], [3, 4]]), [[1, 2], [3, 4]])
        with self.assertRaises(ValidationError):
            func([[1, "2"], [3, 4]])

    def test_set(self):
        @validate
        def set_func(a: set[int]) -> str:
            return str(a)

        self.assertEqual(set_func({1, 2, 3}), "{1, 2, 3}")
        with self.assertRaises(ValidationError):
            set_func({1, "a", 3})

        with self.assertRaises(ValidationError):
            set_func([1, 2, 3])

        @validate(config={"performance": True})
        def set_func2(a: set[int]) -> str:
            return str(a)

        self.assertEqual(set_func2({1, 2, 3}), "{1, 2, 3}")

    def test_typed_dict(self):
        class Person(TypedDict):
            name: str
            age: int

        @validate
        def func(p: Person):
            return p["name"]

        self.assertEqual(func({"name": "Alice", "age": 30}), "Alice")
        with self.assertRaises(ValidationError):
            func({"name": "Alice", "age": "Unknown"})

    def test_literal_types(self):
        @validate
        def func(a: Literal["Yes", "No"]):
            return a

        self.assertEqual(func("Yes"), "Yes")
        self.assertEqual(func("No"), "No")
        with self.assertRaises(ValidationError):
            func("Maybe")

    def test_any(self):
        @validate(config={"strict": False})
        def any_func(a: Any) -> str:
            return str(a)

        self.assertEqual(any_func(5), "5")
        self.assertEqual(any_func("Hello"), "Hello")
        self.assertEqual(any_func([1, 2, 3]), "[1, 2, 3]")

        @validate
        def any_func2(a: Any) -> str:
            return str(a)

        with self.assertRaises(ValidationError):
            any_func2(5)

        any_func2.checking_off()
        self.assertEqual(any_func2(5), "5")

    # def test_async_functions(self):
    #     @validate
    #     async def func(a: int) -> int:
    #         return a

    #     loop = asyncio.get_event_loop()
    #     self.assertEqual(loop.run_until_complete(func(10)), 10)
    #     with self.assertRaises(ValidationError):
    #         loop.run_until_complete(func("test"))



from dataclasses import dataclass, field
from functools import singledispatchmethod
import types, typing, logging
from typing import (
    Any,
    Callable,
)


from ..errors import *
from .struct import GenericBindings
from .config import BindCheckerConfig


log = logging.getLogger(__name__)


class BindChecker:
    """Checks if a value can bind to a type annotation given some already bound states."""

    def __init__(self, config: dict | BindCheckerConfig):
        self.Gbinds = None
        self.config = config

        self.custom_validators = {}

    def new_bindings(self, generics):
        self.Gbinds = GenericBindings(generics)

    def register_validator(self, ty, handler: Callable[[type, Any], None]):
        """Register a handler for a type annotation."""
        self.check.register(ty)(handler)
        log.debug(f"Registered {handler=} for {ty=}")

    def register_custom_validator(self, ty, handler: Callable[[type, Any], None]):
        """Register a handler for a type annotation.

        Could be part of register validator, but there are many catches to custom validation.
            Should only use on primitive types.
        """
        self.custom_validators.setdefault(ty, []).append(handler)
        log.debug(f"Registered custom {handler=} for {ty=}")

    @property
    def checked(self):
        return [val.can_bind_generic for val in self.Gbinds.values()]

    def __check_with_custom_validators(self, ty, value):
        """Check a value against custom validators."""
        if not self.config.use_custom_validators:
            return

        if ty in self.custom_validators:
            for handler in self.custom_validators[ty]:
                valid = handler(value)
                if valid is not None and not valid:
                    raise ValidationError(f"{value=} failed to bind to {ty=}")

    @singledispatchmethod
    def check(
        self,
        ann: int | float | str | bool | bytes | type(None) | type | typing._AnyMeta,
        arg: Any,
    ):
        """Check if a value can bind to a type annotation."""
        log.debug(f"Base: {ann=} {arg=}")

        if type(ann) == typing._AnyMeta:
            if self.config.strict:
                raise ValidationError(
                    f"Type {type(ann)} for `{arg}: {ann=}` must be validated, it cannot be left un-annotated! Disable strict validation to allow this."
                )
            else:
                return

        # if newtype we need to check against the base type
        if hasattr(ann, "__supertype__"):
            ann = ann.__supertype__

        if not isinstance(arg, ann):
            raise InvalidType(f"{ann=} can not validate {arg=}")

        self.__check_with_custom_validators(ann, arg)

    @check.register
    def _(
        self,
        ann: types.GenericAlias | typing._GenericAlias | typing._SpecialGenericAlias,
        arg: Any,
    ):
        log.debug(f"GenericAlias: {ann=} {arg=}")

        if hasattr(ann, "__args__") and len(ann.__args__):
            # TODO: These are really hacky...using {} and []...etc.. :(
            if issubclass(ann.__origin__, dict):
                self.check({"arg_types": ann.__args__}, arg)
            elif issubclass(ann.__origin__, list):
                self.check([*ann.__args__], arg)
            elif issubclass(ann.__origin__, tuple):
                self.check(ann.__args__, arg)
            elif issubclass(ann.__origin__, set):
                self.check(set(ann.__args__), arg)

    @check.register
    def _(self, ann: typing.TypeVar, arg: Any):
        """Handle TypeVars"""
        log.debug(
            f"TypeVar: {ann=} {arg=}\n{ann.__constraints__=}, {type(arg)=}, {type(arg) in [c for c in ann.__constraints__]=}"
        )

        if len(ann.__constraints__):
            constraint_types = [type(c) for c in ann.__constraints__]
            if type(arg) in constraint_types:
                # check against constraints
                self.check(type(arg), arg)
            elif type(arg) not in ann.__constraints__:
                raise ValidationError(
                    f"{arg=} is not valid for {ann=} with constraints {ann.__constraints__}"
                )

        if not self.config.ignore_generics:
            self.Gbinds.try_bind_new_arg(ann, arg)

    @check.register
    def _(self, ann: list, arg: Any):
        """Handle generic sequences"""
        log.debug(f"list: {ann=} {arg=}")

        if not isinstance(arg, list):
            raise InvalidType(f"{arg=} is not a list")

        if self.config.no_list_check or self.config.performance:
            return

        # list like list[T] or list[X]
        if len(ann) == 1:
            for a in arg:
                self.check(ann[0], a)

    @check.register
    def _(self, ann: set, arg: Any):
        """Handle generic sets"""
        log.debug(f"set: {ann=} {arg=}")

        if not isinstance(arg, set):
            raise InvalidType(f"{arg=} is not a set")

        if self.config.no_list_check or self.config.performance:
            return

        # set like set[T] or set[X]
        if len(ann) == 1:
            set_type = next(iter(ann))
            for a in arg:
                self.check(set_type, a)

    @check.register
    def _(self, ann: tuple, arg: Any):
        log.debug(f"tuple: {ann=} {arg=}")

        if not isinstance(arg, tuple):
            raise InvalidType(f"{arg=} is not a tuple")

        if self.config.no_tuple_check or self.config.performance:
            return

        if len(ann) == len(arg):
            # each arg in tuple must bind to each ann in tuple
            for a, b in zip(ann, arg):
                self.check(a, b)

    @check.register
    def _(self, ann: dict, arg: Any):
        log.debug(f"dict: {ann=} {arg=}")

        if not isinstance(arg, dict):
            raise InvalidType(f"{arg=} is not a dict")

        if self.config.no_dict_check or self.config.performance:
            return

        if ann["arg_types"] is not None:
            for k, v in arg.items():
                self.check(ann["arg_types"][0], k)
                self.check(ann["arg_types"][1], v)

    @check.register
    def _(self, ann: types.UnionType | typing._UnionGenericAlias, arg: Any):
        """Handle union types"""
        log.debug(f"Union: {ann=} {arg=}")

        for a in ann.__args__:
            try:
                # TODO: This probably will cause a bug as it could bind and then fail, leaving some bound remnants.
                self.check(a, arg)  # , update_bindings=False) # ?
                return
            except ValidationError:
                pass

        raise ValidationError(f"{arg=} failed to bind to {ann=}")

    @check.register
    def _(self, ann: typing._UnpackGenericAlias, arg: Any):
        """Handle unpacked generic types"""
        log.debug(
            f"UnpackGenericAlias: {ann=} {arg=}, {ann.__origin__=} {ann.__args__=}"
        )

        # support for single type like list[int]
        if len(ann.__args__) == 1:
            log.debug(
                f"UnpackGenericAlias: {self.Gbinds} {ann.__args__[0]=} {type(ann.__args__[0])=}, {type(arg)=}"
            )
            self.Gbinds.try_bind_new_arg(ann.__args__[0], arg)
            log.debug(f"UnpackGenericAlias: {self.Gbinds}")

    @check.register
    def _(self, ann: typing.TypeVarTuple, arg: Any):
        """Handle TypeVarTuples"""
        log.debug(f"TypeVarTuple: {ann=} {arg=}")

        if self.config.no_tuple_check or self.config.performance:
            return

        for e in arg:
            self.Gbinds.try_bind_new_arg(ann, e)

    @check.register
    def _(self, ann: typing._TypedDictMeta, arg: Any):
        """Handle TypedDicts"""
        log.debug(f"TypedDictMeta: {ann=} {arg=}")

        if self.config.no_dict_check or self.config.performance:
            return

        for k, v in arg.items():
            self.check(ann.__annotations__[k], v)

    @check.register
    def _(self, ann: typing._LiteralGenericAlias, arg: Any):
        """Handle Literal types"""
        log.debug(f"LiteralGenericAlias: {ann=} {arg=}")

        if arg not in ann.__args__:
            raise ValidationError(f"{arg=} failed to bind to {ann=}")

    @check.register
    def _(self, ann: typing._CallableGenericAlias, arg: Any):
        """Handle Callable types"""
        log.debug(f"CallableGenericAlias: {ann=} {arg=}")

        if not callable(arg):
            raise ValidationError(f"{arg=} failed to bind to {ann=}")
        else:
            if len(ann.__args__):
                if arg.__name__ == "<lambda>" and not self.config.implied_lambdas:
                    raise ValidationError(
                        f"lambda {arg=} cannot have the required annotations, use a def"
                    )

                if hasattr(arg, "__annotations__"):
                    ann_args = {
                        k: v for (k, v) in arg.__annotations__.items() if k != "return"
                    }
                    ann_ret = arg.__annotations__.get("return")

                    if ann_ret is not None:
                        # Assuming return type is the last in __args__
                        self.check(ann_ret, ann.__args__[-1]())

                    if len(ann_args):
                        for idx, (arg_name, arg_type) in enumerate(ann_args.items()):
                            expected_type = ann.__args__[idx]
                            self.check(expected_type, arg_type())



import logging
from functools import wraps
from typing import (
    Any,
    Callable,
    Optional,
)


from ..errors import *
from ..binding import BindChecker, BindCheckerConfig


log = logging.getLogger(__name__)


class ValidatorFunction(Callable):
    """Callable wrapper for typifying validator functions."""

    def __init__(
        self,
        fn: Callable[..., bool],
        base_type: Optional[type] = None,
        config: Optional[dict] = None,
    ):
        self.fn = fn
        self.base_type = base_type

        self.__call__ = wraps(fn)(self)

        default_cfg = BindCheckerConfig()

        if config is not None:
            default_cfg.update(config)
            if "name" in config:
                self.name = f"_{config['name']}"
                fn.__name__ = self.name
        else:
            self.name = fn.__name__

        self.config = default_cfg

    def __call__(self, value):
        return self.fn(value)

    def set_my_annotations(
        self, annotations: dict, return_annotation=(bool, str | None)
    ):
        self.__annotations__ = annotations
        if return_annotation is not None:
            if not isinstance(return_annotation, tuple):
                self.__annotations__["return"] = return_annotation
            else:
                self.__annotations__["return"] = return_annotation[0]
                if return_annotation[1] is not None:
                    self.config["error"] = return_annotation[1]

    def __and__(self, other: "ValidatorFunction"):
        return ValidatorFunction(lambda value: self.fn(value) and other.fn(value))

    def __or__(self, other: "ValidatorFunction") -> "ValidatorFunction":
        def wrapper(value):
            try:
                result = self.fn(value)
            except ValidationError as e:
                result = False

            result |= other.fn(value)

            return result

        return ValidatorFunction(wrapper)


class ValidationBindChecker(BindChecker):
    def __init__(self, config=None):
        super().__init__(config=config)
        # self.check.register(self.vf_check)
        self.register_validator(ValidatorFunction, self.vf_check)

    def vf_check(self, ann: ValidatorFunction, arg: Any):
        # TODO: Fix this, exceptions r 2 slow, probably.
        # try/except to allow fallback to base_type if VF call fails
        try:
            result = ann(arg)
            if not result:
                error = ann.config["error"]
        except Exception as e:
            result = False
            error = e

        if not result:
            if ann.base_type is not None and not isinstance(arg, ann.base_type):
                raise InvalidType(f"{arg=} is not {ann.base_type=}: {error}")
            elif ann.base_type is None:
                raise ValidationError(f"{arg=} failed validation for {ann=}: {error}")
